module Main (main) where

----
-- protokoll on imporditud qualified võtmesõnaga, et vältida
-- nimekonflikte
import qualified Prot as Prot
import qualified Field as Field

----
-- süsteemi moodulid mida kasutame
import System.Random
import System.IO 
import Control.Exception (finally)
import Network

type Rand a = StdGen -> (a, StdGen)

returnR :: a -> Rand a
returnR x = \gen -> (x, gen)

----
-- Port mida server kasutab.
port :: PortID
port = PortNumber 49267

----
-- Serveri aadress.
host :: HostName
host = "127.0.0.1"

----
-- Loob 50 paarikaupa erinevat väljakut.
mkGames :: Prot.GameConf -> Int -> Rand [Prot.Board]
mkGames cfg seed = 
	returnR (generateGames cfg 50 (Field.randList 50 1 1000 seed))
	
generateGames :: Prot.GameConf -> Int -> [Int] -> [Prot.Board]
generateGames _ 0 _ = [];
generateGames (Prot.GameConf h w m) amount (seed:seeds) =
	-- (Field.randomFieldToInt (Field.myRandomField w h m seed) h h w) : generateGames (Prot.GameConf h w m) (amount-1) seeds
	(Field.randomFieldToInt (Field.myRandomField h w m seed) h h w) : generateGames (Prot.GameConf h w m) (amount-1) seeds 

mkField :: Prot.GameConf -> Field.Field
mkField (Prot.GameConf h w m) =
	Field.emptyField h w m
	
----
-- Väljaku uuendamine.
-- Opened
-- Flagged
updateField :: [Prot.Open] -> [Prot.Coord] -> Field.Field -> Field.Field
updateField opened flagged field =
	Field.update (uniteCells opened flagged) field

uniteCells :: [Prot.Open] -> [Prot.Coord] -> [(Field.Coord, Field.Cell)]
uniteCells [] [] = []
uniteCells [] (f:flagged) =
	(((snd(f)-1), (fst(f)-1)), Field.Flagged) : uniteCells [] flagged
uniteCells (o:opened) [] =
	(((snd(getCoord(o))-1),(fst(getCoord(o))-1)), (Field.Open (getInt(o)))) : uniteCells opened []
uniteCells (o:opened) flagged =
	(((snd(getCoord(o))-1),(fst(getCoord(o))-1)), (Field.Open (getInt(o)))) : uniteCells opened flagged
	
getCoord :: Prot.Open -> (Int, Int)
getCoord (Prot.Open coord n) = coord

getInt :: Prot.Open -> Int
getInt (Prot.Open coord n) = n

----
-- Väljaku lahendamine.
-- NB! Serveri koordinaadid on kujul (y, x), kusjuures (1, 1) on ülemine vasak
-- ning (height, width) parem alumine nurk.

type Row = [Int]
type Matrix = [Row]

----
-- Solver + Tuvastaja
solveField :: Field.Field -> [Prot.Board] -> Rand ([Prot.Coord], [Prot.Coord])
solveField field games
	| board /= Nothing = returnR (boardToCoords (maybeBoardToBoard board) field)
	| otherwise = returnR (solve field (makeEquationMatrix field (getOpenCells field)) (equationVariablesList field))
	where
		board = guessField games field

{-
---- Solver only
solveField :: Field.Field -> [Prot.Board] -> Rand ([Prot.Coord], [Prot.Coord])
solveField field games
	= returnR (solve field (makeEquationMatrix field (getOpenCells field)) (equationVariablesList field))
-}



getAnyClosedCell :: Field.Field -> Prot.Coord
getAnyClosedCell (f:field)
	| (snd(f) == Field.Closed) || (snd(f) == Field.Mined) = fieldCoordToProtCoord (fst(f))
	| otherwise = getAnyClosedCell field

----
-- Get list of open cells, to create equations for
----
getOpenCells :: Field.Field -> [(Field.Coord,Int)]
getOpenCells [] = []
getOpenCells (item:restOfField)
	| cellClosedOrOpen (snd (item)) == -1 = getOpenCells restOfField
	| otherwise = (fst(item), (cellClosedOrOpen (snd (item)))) : getOpenCells restOfField

cellClosedOrOpen :: Field.Cell -> Int
cellClosedOrOpen (Field.Open x) = x
cellClosedOrOpen _ = -1



---------------------------------------------
-- Represent as equations
---------------------------------------------

----
-- For each open cell get surrounding cells and build equation matrix
-- Closed will be 1 in equation
-- Flagged will be -1 in equation (?? or should it be 0, but result - 1 ??)
-- Every other will be 0 in equation
----

makeEquationMatrix :: Field.Field -> [(Field.Coord, Int)] -> Matrix
makeEquationMatrix field open
	= updateMatrix (makeEquationMatrix' field open)

makeEquationMatrix' :: Field.Field -> [(Field.Coord, Int)] -> Matrix
makeEquationMatrix' _ [] = []
makeEquationMatrix' field (o:open)
	= (makeEquationRow (equationVariablesList field) (getSortedEquationRow (Field.surroundingCells field (fst(o)))) (snd(o))) : makeEquationMatrix' field open

{-
----
-- Unoptimized
updateMatrix :: Matrix -> Matrix
updateMatrix [] = []
updateMatrix (m:matrix)
	= (updateRow m) : updateMatrix matrix
-}

----
-- Optimized
updateMatrix :: Matrix -> Matrix
updateMatrix [] = []
updateMatrix (m:matrix)
	| sum um == 0 = updateMatrix matrix
	| otherwise = um : updateMatrix matrix
	where
		um = updateRow m


updateRow :: Row -> Row
updateRow row
	= (updateFlaggedInRow (listWithoutLast row)) ++ [(last row + sum (filter (==(-1)) row))]

updateFlaggedInRow :: Row -> Row
updateFlaggedInRow [] = []
updateFlaggedInRow (r:row)
	| r == (-1) = 0 : updateFlaggedInRow row
	| otherwise = r : updateFlaggedInRow row

makeEquationRow :: [Field.Coord] -> [(Field.Coord, Int)] -> Int -> Row
makeEquationRow [] _ result = result:[]
makeEquationRow (_:places) [] result
	= 0 : makeEquationRow places [] result
makeEquationRow (p:places) (x:xs) result
	| p == fst(x) = snd(x) : makeEquationRow places xs result
	| otherwise = 0 : makeEquationRow places (x:xs) result

{-
prepareEquationData :: [(Field.Coord, Int)] -> Field.Field -> [([(Field.Coord,Int)], Int)]
prepareEquationData (o:open) field
	= undefined
-}

-- x1, x2, x3, ..., xn
equationVariablesList :: Field.Field -> [(Field.Coord)]
equationVariablesList field
	= quicksort (unique [(x,y) | ((x,y),z) <- mapCellTypeWithEquation (allCellsInEquation field (getOpenCells field))])
	
allCellsInEquation :: Field.Field -> [(Field.Coord, Int)] -> [(Field.Coord, Field.Cell)]
allCellsInEquation _ [] = []
allCellsInEquation field (r:rest)
	= (Field.surroundingCells field (fst(r))) ++ allCellsInEquation field rest

getSortedEquationRow :: [(Field.Coord, Field.Cell)] -> [(Field.Coord, Int)]
getSortedEquationRow list
	= quicksort (mapCellTypeWithEquation (list))

quicksort [] = []
quicksort (x:xs) = quicksort small ++ mid ++ quicksort large
	where
		small = [y | y<-xs, y<x]
		mid = [y | y<-xs, y==x] ++ [x]
		large = [y | y<-xs, y>x]	
		
unique [] = []
unique (x:xs)
	| elem x xs = unique xs
	| otherwise = x : unique xs

mapCellTypeWithEquation :: [(Field.Coord, Field.Cell)] -> [(Field.Coord, Int)]
mapCellTypeWithEquation [] = []
mapCellTypeWithEquation ((coord,cell):rest)
	| cell == Field.Closed = (coord, 1) : mapCellTypeWithEquation rest
	| cell == Field.Mined = (coord, 1) : mapCellTypeWithEquation rest
	| cell == Field.Flagged = (coord, (-1)) : mapCellTypeWithEquation rest
	| otherwise = (coord, 0) : mapCellTypeWithEquation rest



---------------------------------------------
-- Solver
---------------------------------------------

solve :: Field.Field -> Matrix -> [Field.Coord] -> ([Prot.Coord], [Prot.Coord])
solve field matrix mapping
	| length(open) > 0 = (open, flagged)
	| otherwise = ([(getAnyClosedCell field)], flagged)
	where 
		(open, flagged) = (unique (findOpen matrix mapping), unique (findMined matrix mapping))

findMined :: Matrix -> [Field.Coord] -> [Prot.Coord]
findMined [] _ = []
findMined (r:rows) mapping
	= (minedRules r mapping) ++ (applyRowSearchMined r rows mapping) ++ (applyRowSearchMixedMined r rows mapping) ++ findMined rows mapping
	
findOpen :: Matrix -> [Field.Coord] -> [Prot.Coord]
findOpen [] _ = []
findOpen (r:rows) mapping
	= (openRules r mapping) ++ (applyRowSearchOpen r rows mapping) ++ (applyRowSearchMixedOpen r rows mapping) ++ findOpen rows mapping

applyRowSearchMined :: Row -> Matrix -> [Field.Coord] -> [Prot.Coord]
applyRowSearchMined _ [] _ = []
applyRowSearchMined row (m:matrix) mapping
	= (minedRules (substractRow row m) mapping) ++ applyRowSearchMined row matrix mapping

applyRowSearchOpen :: Row -> Matrix -> [Field.Coord] -> [Prot.Coord]
applyRowSearchOpen _ [] _ = []
applyRowSearchOpen row (m:matrix) mapping
	= (openRules (substractRow row m) mapping) ++ applyRowSearchOpen row matrix mapping

applyRowSearchMixedOpen :: Row -> Matrix -> [Field.Coord] -> [Prot.Coord]
applyRowSearchMixedOpen _ [] _ = []
applyRowSearchMixedOpen row (m:matrix) mapping
	= (mixedRulesOpen (substractRow row m) mapping) ++ applyRowSearchMixedOpen row matrix mapping

applyRowSearchMixedMined :: Row -> Matrix -> [Field.Coord] -> [Prot.Coord]
applyRowSearchMixedMined _ [] _ = []
applyRowSearchMixedMined row (m:matrix) mapping
	= (mixedRulesMined (substractRow row m) mapping) ++ applyRowSearchMixedMined row matrix mapping

substractRow :: Row -> Row -> Row
substractRow [] [] = []
substractRow (m:minuend) (s:subtrahend) =
	(m-s) : substractRow minuend subtrahend
	
----
-- Cases then 1 or -1 indicates mine
-- [1,1,0,0,1,3]
-- [-1,-1,0,0,0,-3]
-- [1,0,0,1,-1,-1,0]
----
minedRules :: Row -> [Field.Coord] -> [Prot.Coord]
minedRules row mapping
	| (sum (listWithoutLast row) == last row) &&
	  ((sum (filter (==(-1)) (listWithoutLast row)) == sum (listWithoutLast row)) ||
	   (sum (filter (==1) (listWithoutLast row)) == sum (listWithoutLast row))) = mapCellsInRow (listWithoutLast row) mapping 1
	| otherwise = []

----
-- Cases then 1 or -1 indicates safe cell
-- [1,0,1,0,0]
-- [-1,0,0,0,0]
----
openRules :: Row -> [Field.Coord] -> [Prot.Coord]
openRules row mapping
	| (last row == 0) && (sum (listWithoutLast row) /= last row) && ((sum (filter (==(-1)) row) == sum row) || (sum (filter (==1) row) == sum row)) = mapCellsInRow (listWithoutLast row) mapping 1
	| otherwise = []

----
-- Cases then 1 indicats mine and -1 safe cell
-- [1,1,0,0,-1,0,2]
-- Cases then -1 indicats mine and 1 safe cell
-- [1,1,0,0,-1,0,-1]
----
mixedRulesMined :: Row -> [Field.Coord] -> [Prot.Coord]
mixedRulesMined row mapping
	| sum (filter (==(-1)) (listWithoutLast row)) == last row = mapNegativeCellsInRow (listWithoutLast row) mapping 1
	| sum (filter (==(1)) (listWithoutLast row)) == last row = mapPositiveCellsInRow (listWithoutLast row) mapping 1
	| otherwise = []
	
mixedRulesOpen :: Row -> [Field.Coord] -> [Prot.Coord]
mixedRulesOpen row mapping
	| sum (filter (==(-1)) (listWithoutLast row)) == last row = mapPositiveCellsInRow (listWithoutLast row) mapping 1
	| sum (filter (==(1)) (listWithoutLast row)) == last row = mapNegativeCellsInRow (listWithoutLast row) mapping 1
	| otherwise = []


listWithoutLast :: Row -> Row
listWithoutLast row = reverse(tail(reverse row))

getCoordFromMapping :: Int -> [Field.Coord] -> Prot.Coord
getCoordFromMapping n list
	= fieldCoordToProtCoord (head (drop (n-1) list))

mapCellsInRow :: Row -> [Field.Coord] -> Int -> [Prot.Coord]
mapCellsInRow [] _ _ = []
mapCellsInRow (r:row) mapping counter
	| (r == 0) = mapCellsInRow row mapping (counter+1)
	| otherwise = (getCoordFromMapping counter mapping) : mapCellsInRow row mapping (counter+1)
	
mapPositiveCellsInRow :: Row -> [Field.Coord] -> Int -> [Prot.Coord]
mapPositiveCellsInRow [] _ _ = []
mapPositiveCellsInRow (r:row) mapping counter
	| (r == 1) = (getCoordFromMapping counter mapping) : mapPositiveCellsInRow row mapping (counter+1)
	| otherwise = mapPositiveCellsInRow row mapping (counter+1)

mapNegativeCellsInRow :: Row -> [Field.Coord] -> Int -> [Prot.Coord]
mapNegativeCellsInRow [] _ _ = []
mapNegativeCellsInRow (r:row) mapping counter
	| (r == (-1)) = (getCoordFromMapping counter mapping) : mapNegativeCellsInRow row mapping (counter+1)
	| otherwise = mapNegativeCellsInRow row mapping (counter+1)
	

fieldCoordToProtCoord :: Field.Coord -> Prot.Coord
fieldCoordToProtCoord (x,y) = (y+1, x+1)

protCoordToFieldCoord :: Prot.Coord -> Field.Coord
protCoordToFieldCoord (x,y) = (y-1, x-1)



---------------------------------------------
-- Tuvastaja
---------------------------------------------

guessField :: [Prot.Board] -> Field.Field -> Maybe Prot.Board
guessField haystack needle
	| length similarFields == 1 = Just (head similarFields)
	| otherwise = Nothing
	where
		similarFields = findSimilarFields haystack needle
	
----
-- Cast Maybe Board to Board
maybeBoardToBoard :: Maybe Prot.Board -> Prot.Board
maybeBoardToBoard (Just board) = board
maybeBoardToBoard Nothing = []
	
----
-- Cast Board to ([open], [put flag])
boardToCoords :: Prot.Board -> Field.Field -> ([Prot.Coord],[Prot.Coord])
boardToCoords board alreadyRevealed =
	((getOpenCoords field), (getMinedCoords field))
	where
		field = onlyNewCells (protBoardToField board 0) alreadyRevealed

----
-- Do not inlude into update message cell which are alredy reveaeled
-- Field1 - full board with everythig open
-- Field2 - current position
-- Return Field = Field1 - Field2
onlyNewCells :: Field.Field -> Field.Field -> Field.Field
onlyNewCells [] [] = []
onlyNewCells (f1:field1) (f2:field2)
	| ((snd f2) == Field.Closed) || ((snd f2) == Field.Mined) = f1 : onlyNewCells field1 field2
	| otherwise = onlyNewCells field1 field2


getMinedCoords :: Field.Field -> [Prot.Coord]
getMinedCoords [] = []
getMinedCoords (f:field)
	| (snd(f) == Field.Mined) = (fieldCoordToProtCoord (fst f)) : getMinedCoords field
	| otherwise = getMinedCoords field

getOpenCoords :: Field.Field -> [Prot.Coord]
getOpenCoords [] = []
getOpenCoords (f:field)
	| (snd(f) == Field.Closed) = (fieldCoordToProtCoord (fst f)) : getOpenCoords field
	| otherwise = getOpenCoords field

----
-- Find fields which are 30%+ similar to current position
findSimilarFields :: [Prot.Board] -> Field.Field -> [Prot.Board]
findSimilarFields [] _ = []
findSimilarFields (h:haystack) needle
	| round ((realToFrac (compareFields (protBoardToField h 0) needle)) / (realToFrac (length needle)) * 100) > 30 = h : findSimilarFields haystack needle
	| otherwise = findSimilarFields haystack needle

protBoardToField :: Prot.Board -> Int -> Field.Field
protBoardToField [] _ = []
protBoardToField (r:row) y
	= (protRowToField r 0 y) ++ protBoardToField row (y+1);

protRowToField :: [Int] -> Int -> Int -> Field.Field
protRowToField [] _ _ = []
protRowToField (c:cells) x y
	| c == 1 = ((x,y), Field.Mined) : protRowToField cells (x+1) y
	| c == 0 = ((x,y), Field.Closed) : protRowToField cells (x+1) y

----
-- If fields are similar (see similarity rules) count number revealed (Open, Flagged) cells, otherwise return 0
-- field1 - Generated one
-- field2 - Current game position
compareFields :: Field.Field -> Field.Field -> Int
compareFields field1 field2
	| (similarityRule1 field1 field2) && (similarityRule2 field1 field2) =
		(length [((x,y),Field.Flagged) | ((x,y),Field.Flagged) <- field2]) + (length [((x,y),(Field.Open n)) | ((x,y),(Field.Open n)) <- field2])
	| otherwise = 0

----
-- Similarity rule 1:
-- a) No cell on field1 is mined when it is opened on field2
-- b) No cell on field1 is not mined when it is flagged on field2
similarityRule1 :: Field.Field -> Field.Field -> Bool
similarityRule1 [] [] = True
similarityRule1 ((_,cell1):field1) ((_,cell2):field2)
	| (cell1 == Field.Mined) && (Field.getOpenCellInt(cell2) > 0) = False
	| (cell1 /= Field.Mined) && (cell2 == Field.Flagged) = False
	| otherwise = similarityRule1 field1 field2

----
-- Similarity rule 2:
-- a) Every open cell on field2 will touch correct number of mines when it is placed on field1
similarityRule2 :: Field.Field -> Field.Field -> Bool
similarityRule2 _ [] = True
similarityRule2 field1 ((coord2,cell2):field2)
	| (Field.getOpenCellInt(cell2) > 0) &&
	  ((length [((x,y),Field.Mined) | ((x,y),Field.Mined) <- (Field.surroundingCells field1 coord2)]) == Field.getOpenCellInt(cell2)) = similarityRule2 field1 field2
	| (Field.getOpenCellInt(cell2) > 0) = False
	| otherwise = similarityRule2 field1 field2

----
-- Protseduur, mis realiseerib serveriga suhtluse.
game :: Handle -> IO ()
game handle = do
  let
    mainLoop cfg games = do
      msg <- recv
      case msg of
        Prot.NewGame name newCfg -> do
          -- putStr $ "Generating games ..."
          n <- getStdRandom (randomR (1::Int,1000))
          games <- getStdRandom (mkGames newCfg n)
          send (Prot.Games games)
          -- putStrLn $ " done: " ++ (show (length games)) ++ " games generated and sent to server"
          mainLoop newCfg games
        Prot.NewField -> do
          roundLoop (mkField cfg) (mainLoop cfg games) games
        Prot.GameOver status -> do
          putStrLn $ "Game over: " ++ show status
          mainLoop (error "Server bug!") games
        Prot.FieldDone score -> do
          putStrLn $ "Solved " ++ show (score*100.0) ++ "% of field."
          -- putStrLn $ "--------------------------------------------------------------------------------"
          mainLoop (error "Server bug!") games
        Prot.End stat -> do
          print stat

    roundLoop field k games = do
      (cs, fs) <- getStdRandom (solveField field games)
      -- Field.printField field
      -- putStrLn $ "Lets try to open " ++ (show cs) ++ " and flag " ++ (show fs)
      send (Prot.Clicks cs fs)
      msg <- recv
      case msg of
        Prot.Opened xs -> do
          -- putStrLn $ "Answer: open " ++ (show xs) ++ " flagged " ++ (show fs) 
          roundLoop (updateField xs fs field) k games
        -- Prot.RoundOver score -> do
        Prot.FieldDone score -> do
          putStrLn $ "Solved " ++ show (score*100.0) ++ "% of field."
          -- putStrLn $ "--------------------------------------------------------------------------------"
          k
        _ -> error "Server bug!"

    send msg = do
      hPutStrLn handle (show msg)
      hFlush handle

    recv = do
      str <- hGetLine handle
      return $ read str
  n <- getStdRandom (randomR (1::Int,1000))
  send (Prot.Hello ("Player " ++ (show n)))
  mainLoop (error "Server bug!") []

main :: IO ()
main = withSocketsDo $ do
  handle <- connectTo host port -- loome ühenduse
  hSetBuffering handle LineBuffering
  (game handle `finally` hClose handle)
			