import Prot
import Field
import ClientThread
import ThreadUtil
import Swiss

import Data.List (sortBy)
import Data.Function (on)
import Data.Map (Map)
import Data.Set (Set)
import qualified Data.Map as M
import qualified Data.Set as S

import System.Random
import System.Environment
import System.IO
import Control.Concurrent.MVar
import Control.Concurrent
import Control.Monad
import Control.Applicative
import Control.Monad.Reader
import Control.Monad.State
import Network
import Control.Exception (bracket, finally)

defaultConf :: GameConf
defaultConf = GameConf {
    height = 7, -- 16 *See below as well
    width = 12, -- 30
    nMines = 20 -- 99
  }

----
-- Client State
----

data Client = Client {
    cid :: Int,
    input :: Chan ClientMessage,
    output :: Chan ServerMessage,
    name :: MVar Name
  }

instance Eq Client where
  c == c' = cid c == cid c'

instance Ord Client where
  compare c c' = compare (cid c) (cid c')

mkClient :: Int ->
	    Chan ClientMessage ->
	    Chan ServerMessage ->
	    IO Client
mkClient i cc sc = do
  clientName <- newMVar "noname"
  return $ Client i cc sc clientName

----
-- Server State
----

-- Ma ei ole kindel, kas mvar on siin kanali salvestamiseks vajalik.

data Server = Server {
    clients :: [Client],
    logHandle :: MVar Handle,
    gen :: MVar StdGen
  }

mkServer :: Handle -> [Client] -> IO Server
mkServer hndl cs = do
  mhandle <- newMVar hndl
  mgen <- newStdGen >>= newMVar
  return $ Server cs mhandle mgen

type ServerM = ReaderT Server IO

serverLoop :: ServerM ()
serverLoop = do
  askForNames
  logLn "Starting tournament..."
  st <- ask
  order <- liftIO (tournament (\l r -> runReaderT (game l r) st) (clients st))
  logLn "Kicking everyone..."
  kickEveryone order

game :: Client -> Client -> ServerM Ordering
game cA cB = do
  aName <- getClientName cA
  bName <- getClientName cB
  logLn "Asking for games..."
  sendMsg cA (NewGame bName defaultConf)
  Games gamesA <- recvMsg cA
  sendMsg cB (NewGame aName defaultConf)
  Games gamesB <- recvMsg cB
  logLn $ "Playing \"" ++ aName ++ "\" and \"" ++ bName ++ "..."
  (toA, toB) <- shuffleGames gamesA gamesB
  [scoreA, scoreB] <- forkAndJoin [playAll cA toA, playAll cB toB]
  case compare scoreA scoreB of
    LT -> do
      sendMsg cA (GameOver Loss)
      sendMsg cB (GameOver Win)
      return LT
    EQ -> do
      sendMsg cA (GameOver Tie)
      sendMsg cB (GameOver Tie)
      return EQ
    GT -> do
      sendMsg cA (GameOver Win)
      sendMsg cB (GameOver Loss)
      return GT

play :: Client -> (Bool, Board) -> ServerM Double
play client (isOwner, board) = case buildField board 7 12 20 of
  Just field -> do
    score <- loop NewField field
    sendMsg client (FieldDone score)
    return score
  Nothing -> return $ if isOwner then 0 else 1
  where
    loop msg field = case gameState field of
      GameDunno -> do
	sendMsg client msg
	Clicks l r <- recvMsg client
	if null l && null r
	  then do
	    logLn "Client messed up (i think):"
	    logLn (showField field "")
	    return (-1.0)
	  else let 
	      (field', xs) = clicks l $ flags r field
	    in loop (Opened xs) field'
      GameWon -> return 1.0
      GameLost -> return $ countCorrect field

playAll :: Client -> [(Bool, Board)] -> ServerM Double
playAll client xs = sum <$> mapM (play client) xs

-- A gets half of B games and half of A games in random order
shuffleGames :: [Board] -> [Board] -> ServerM ([(Bool, Board)], [(Bool, Board)])
shuffleGames gamesA gamesB = do
  (a1, b1) <- splitAt (n `div` 2) <$> shuffle gamesA
  (b2, a2) <- splitAt (m `div` 2) <$> shuffle gamesB
  toA <- shuffle (tify a1 ++ fify a2)
  toB <- shuffle (tify b1 ++ fify b2)
  return (toA, toB)
  where
    n = length gamesA
    m = length gamesB
    tify = map (\x -> (True, x))
    fify = map (\x -> (False, x))

-- could be better
shuffle :: [a] -> ServerM [a]
shuffle xs = do
  gen <- getGen
  return . map fst . sortBy (compare `on` snd) $ zip xs (randoms gen :: [Int])

askForNames :: ServerM ()
askForNames = do
  logLn "Asking for names..."
  forClients_ $ \client -> do
    Hello newName <- recvMsg client 
    liftIO $ modifyMVar_ (name client) (\_ -> return newName)

kickEveryone :: [[Client]] -> ServerM ()
kickEveryone = loop 1
  where
    loop n [] = return ()
    loop n (group : groups) = do
      let m = length group
      forM_ group $ \client -> do
	let msg = End (TournamentOver { position = (n, n + m - 1) } )
	sendMsg client msg
      loop (n + m) groups

logLn :: String -> ServerM ()
logLn msg = do
  mhandle <- asks logHandle
  liftIO $ do
    handle <- takeMVar mhandle
    hPutStrLn handle msg
    hFlush handle
    putMVar mhandle handle

recvMsg :: Client -> ServerM ClientMessage
recvMsg = liftIO . readChan . input

sendMsg :: Client -> ServerMessage -> ServerM ()
sendMsg client = liftIO . writeChan (output client)

getClientName :: Client -> ServerM Name
getClientName client = liftIO $ readMVar $ name client

forClients_ :: (Client -> ServerM ()) -> ServerM ()
forClients_ act = do
  cs <- asks clients
  forM_ cs act

getGen :: ServerM StdGen
getGen = do
  mgen <- asks gen 
  liftIO $ modifyMVar mgen (return . split)

main :: IO ()
main = withSocketsDo $ do
  args <- getArgs
  case args of
    [n_str] -> case reads n_str of
      [(n, [])] | n > 1 -> startGame n
      _ -> error failMsg
    _ -> error failMsg
  where
    failMsg = "Ainult üks naturaalarvuline parameeter, palun."

startGame :: Int -> IO ()
startGame n = listenOrClose $ \soc -> do
  children <- newMVar []
  logHndl <- openFile "log.txt" WriteMode
  let fork = forkChild children
  xs <- forM [1 .. n] $ \i -> do
    (handle, host, _) <- accept soc
    hSetBuffering handle LineBuffering
    inChan <- newChan
    outChan <- newChan
    fork (initClientThread handle inChan outChan)
    return (i, inChan, outChan)
  fork (initServerThread logHndl xs) 
  waitForChildren children
  hClose logHndl

initServerThread :: Handle ->
                    [(Int, Chan ClientMessage, Chan ServerMessage)] -> IO ()
initServerThread hndl xs = do
  cs <- forM xs $ \(i, inp, out) -> mkClient i inp out
  server <- mkServer hndl cs
  runReaderT serverLoop server

port :: PortID
port = PortNumber 49267

listenOrClose :: (Socket -> IO a) -> IO a
listenOrClose = bracket (listenOn port) sClose

----
-- Jah, see on kaunis kole...
----

forkAndJoin :: [ServerM a] -> ServerM [a]
forkAndJoin acts = do
  st <- ask
  liftIO $ do
    mxs <- forM acts $ \act -> do
      mx <- newEmptyMVar
      forkIO $ do
	x <- runReaderT act st
	putMVar mx x
      return mx
    mapM takeMVar mxs
